#!/usr/bin/env python3
#
#  parser.py
"""
Configuration parser.
"""
#
#  Copyright © 2020-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import tempfile
from typing import Any, List, Mapping, MutableMapping

# 3rd party
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike
from ruamel.yaml import YAML

# this package
from configconfig.metaclass import ConfigVarMeta
from configconfig.utils import make_schema
from configconfig.validator import validate_files

__all__ = ["Parser"]


class Parser:
	"""
	Base class for YAML configuration parsers.

	Custom parsing steps for each configuration variable can be implemented with methods in the form:

	.. code-block:: python

		def visit_<configuration value name>(
				self,
				raw_config_vars: Dict[str, Any],
				) -> Any: ...

	The method must return the value to set the configuration variable to,
	or raise an error in the case of invalid input.

	|

	.. latex:vspace:: -55px

	A final custom parsing step, useful when several values must be set at once,
	may be implemented in the ``custom_parsing`` method:

	.. code-block:: python

		def custom_parsing(
				self,
				raw_config_vars: Mapping[str, Any],
				parsed_config_vars: MutableMapping[str, Any],
				filename: PathPlus,
				) -> MutableMapping[str, Any]: ...

	This takes the mapping of raw configuration variables,
	the mapping of parsed variables (those set with the ``visit_<configuration value name>`` method),
	and the configuration file name. The method must return the ``parsed_config_vars``.
	"""

	config_vars: List[ConfigVarMeta]

	def __init__(self, allow_unknown_keys: bool = False):
		self.allow_unknown_keys = allow_unknown_keys

	def run(self, filename: PathLike):
		"""
		Parse configuration from the given file.

		:param filename: The filename of the YAML configuration file.
		"""

		filename = PathPlus(filename)

		if not filename.is_file():
			raise FileNotFoundError(str(filename))

		with tempfile.TemporaryDirectory() as tmpdir:
			tmpdir_p = PathPlus(tmpdir)
			schema_file = tmpdir_p / "schema.json"
			schema = make_schema(*self.config_vars)
			schema["additionalProperties"] = self.allow_unknown_keys
			schema_file.dump_json(schema)
			validate_files(schema_file, filename)

		parsed_config_vars: MutableMapping[str, Any] = {}

		with filename.open() as file:
			raw_config_vars: Mapping[str, Any] = YAML(typ="safe", pure=True).load(file)

		for var in self.config_vars:
			parsed_config_vars[var.__name__] = getattr(self, f"visit_{var.__name__}", var.get)(raw_config_vars)

		return self.custom_parsing(raw_config_vars, parsed_config_vars, filename)

	def custom_parsing(
			self,
			raw_config_vars: Mapping[str, Any],
			parsed_config_vars: MutableMapping[str, Any],
			filename: PathPlus,
			) -> MutableMapping[str, Any]:
		"""
		Custom parsing step.

		:param raw_config_vars: Mapping of raw configuration values loaded from the YAML configuration file.
		:param parsed_config_vars: Mapping of parsed configuration values.
		:param filename: The filename of the YAML configuration file.
		"""

		return parsed_config_vars



#!/usr/bin/env python3
#
#  configvar.py
"""
Base class for ``YAML`` configuration values.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from textwrap import dedent, indent
from typing import Any, Callable, Optional, Type, Union

# 3rd party
from domdf_python_tools.stringlist import StringList
from typing_inspect import is_literal_type  # type: ignore

# this package
from configconfig.metaclass import ConfigVarMeta
from configconfig.utils import RawConfigVarsType, get_yaml_type, tab
from configconfig.validator import Validator

__all__ = ["ConfigVar"]


class ConfigVar(metaclass=ConfigVarMeta):
	"""
	Base class for ``YAML`` configuration values.

	The class docstring should be the description of the config var, with an example,
	and the name of the class should be the variable name.

	If you would prefer a more Pythonic naming approach the variable name can
	be configured with the ``name`` class variable.

	.. latex:vspace:: -5px

	:bold-title:`Example:`

	.. code-block:: python

		class platforms(ConfigVar):
			\"\"\"
			A case-insensitive list of platforms to perform tests for.

			Example:

			.. code-block:: yaml

				platforms:
				  - Windows
				  - Linux

			These values determine the GitHub test workflows to enable,
			and the Trove classifiers used on PyPI.
			\"\"\"

			dtype = List[Literal["Windows", "macOS", "Linux"]]
			default: List[str] = ["Windows", "macOS", "Linux"]
			category: str = "packaging"

	.. latex:vspace:: -10px

	"""  # noqa: D300,D301

	dtype: Type
	"""
	The allowed type or types in the ``YAML`` configuration file.
	"""

	rtype: Type
	"""
	The variable type passed to Jinja2.
	If ``None`` :attr:`~configconfig.configvar.ConfigVar.dtype` is used.
	Ignored for ``dtype=bool``.
	"""

	required: bool
	"""
	Flag to indicate whether the configuration value is required. Default :py:obj:`False`.
	"""

	default: Union[Callable[[RawConfigVarsType], Any], Any]
	"""
	The default value of the configuration value if it is optional. Defaults to ``''`` if unset.

	May also be set to a callable which returns a dynamic or mutable default value.
	"""

	category: str
	"""
	The category the :class:`~configconfig.configvar.ConfigVar` is listed under in the documentation.
	"""

	__name__: str

	@classmethod
	def validator(cls, value: Any) -> Any:
		"""
		Function to call to validate the values.

		* The callable must have a single required argument (the value).
		* Should raise :exc:`ValueError` if values are invalid, and return the values if they are valid.
		* May change the values (e.g. make lowercase) before returning.
		"""

		return value

	def __new__(cls, raw_config_vars: RawConfigVarsType) -> Any:  # noqa: D102
		# Exists purely so mypy knows about the signature
		return cls.get(raw_config_vars)  # pragma: no cover

	@classmethod
	def get(cls, raw_config_vars: Optional[RawConfigVarsType] = None) -> Any:
		"""
		Returns the value of this :class:`~configconfig.configvar.ConfigVar`.

		:param raw_config_vars: Dictionary to obtain the value from.

		:rtype: See the :attr:`~.ConfigVar.rtype` attribute.
		"""

		return cls.validator(cls.validate(raw_config_vars))

	@classmethod
	def validate(cls, raw_config_vars: Optional[RawConfigVarsType] = None) -> Any:
		"""
		Validate the value obtained from the ``YAML`` file and coerce into the appropriate return type.

		:param raw_config_vars: Dictionary to obtain the value from.

		:rtype: See the :attr:`~.ConfigVar.rtype` attribute.
		"""

		if raw_config_vars is None:
			raw_config_vars = {}

		if cls.rtype is None:
			cls.rtype = cls.dtype

		validator = Validator(cls)
		return validator.validate(raw_config_vars)

	@classmethod
	def make_documentation(cls):
		"""
		Returns the reStructuredText documentation for the :class:`~.ConfigVar`.
		"""

		docstring = cls.__doc__ or ''
		docstring = (indent(dedent(docstring), tab))

		if not docstring.startswith('\n'):
			docstring = '\n' + docstring

		buf = StringList()
		buf.indent_type = "    "
		buf.blankline(ensure_single=True)
		buf.append(f".. conf:: {cls.__name__}")
		buf.append(docstring)
		buf.blankline()

		buf.indent_size += 1

		buf.append(f"**Required**: {'yes' if cls.required else 'no'}")
		buf.blankline()
		buf.blankline()

		if not cls.required:
			if cls.default == []:
				buf.append("**Default**: [ ]")
			elif cls.default == {}:
				buf.append("**Default**: { }")
			elif isinstance(cls.default, Callable):  # type: ignore
				buf.append(f"**Default**: The value of :conf:`{cls.default.__name__}`")
			elif isinstance(cls.default, bool):
				buf.append(f"**Default**: :py:obj:`{cls.default}`")
			elif isinstance(cls.default, str):
				if cls.default == '':
					buf.append("**Default**: <blank>")
				else:
					buf.append(f"**Default**: ``{cls.default}``")
			else:
				buf.append(f"**Default**: {cls.default}")

			buf.blankline()
			buf.blankline()

		buf.append(f"**Type**: {get_yaml_type(cls.dtype)}")

		if is_literal_type(cls.dtype):
			valid_values = ", ".join(f"``{x}``" for x in cls.dtype.__args__)
			buf.blankline()
			buf.blankline()
			buf.append(f"**Allowed values**: {valid_values}")
		elif hasattr(cls.dtype, "__args__") and is_literal_type(cls.dtype.__args__[0]):
			valid_values = ", ".join(f"``{x}``" for x in cls.dtype.__args__[0].__args__)
			buf.blankline()
			buf.blankline()
			buf.append(f"**Allowed values**: {valid_values}")

		buf.indent_size -= 1

		return str(buf)



#!/usr/bin/env python3
#
#  utils.py
"""
Utility functions.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  gat_args based on CPython.
#  Licensed under the Python Software Foundation License Version 2.
#  Copyright © 2001-2020 Python Software Foundation. All rights reserved.
#  Copyright © 2000 BeOpen.com . All rights reserved.
#  Copyright © 1995-2000 Corporation for National Research Initiatives . All rights reserved.
#  Copyright © 1991-1995 Stichting Mathematisch Centrum . All rights reserved.
#

# stdlib
import copy
import sys
import typing
from enum import EnumMeta
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Type, TypeVar, Union

# 3rd party
from typing_extensions import Literal
from typing_inspect import is_literal_type  # type: ignore

if TYPE_CHECKING:
	# this package
	from configconfig.metaclass import ConfigVarMeta

if sys.version_info >= (3, 8):  # pragma: no cover (<py38)

	# stdlib
	from typing import get_args, get_origin

else:  # pragma: no cover (>=py38)

	# stdlib
	import collections.abc

	# 3rd party
	from typing_inspect import get_origin

	def get_args(tp):
		"""Get type arguments with all substitutions performed.

		For unions, basic simplifications used by Union constructor are performed.

		Examples::

			get_args(Dict[str, int]) == (str, int)
			get_args(int) == ()
			get_args(Union[int, Union[T, int], str][int]) == (int, str)
			get_args(Union[int, Tuple[T, int]][str]) == (int, Tuple[str, int])
			get_args(Callable[[], T][int]) == ([], int)
		"""
		if hasattr(tp, "__args__"):
			res = tp.__args__
			if get_origin(tp) is collections.abc.Callable and res[0] is not Ellipsis:
				res = (list(res[:-1]), res[-1])
			return res
		return ()


__all__ = [
		"get_literal_values",
		"basic_schema",
		"optional_getter",
		"get_yaml_type",
		"make_schema",
		"check_union",
		"get_json_type",
		"RawConfigVarsType",
		]

#: A literal ``TAB`` (``\t``) character for use in f-strings.
tab = '\t'

if sys.version_info < (3, 7):
	UnionType = Any
	GenericAliasType = Any
else:
	UnionType = type(Union)
	GenericAliasType = type(List)


def optional_getter(raw_config_vars: Dict[str, Any], cls: "ConfigVarMeta", required: bool) -> Any:
	"""
	Returns either the configuration value, the default,
	or raises an error if the value is required but wasn't supplied.

	:param raw_config_vars:
	:param cls:
	:param required:
	"""  # noqa: D400

	if required:
		try:
			return raw_config_vars[cls.__name__]
		except KeyError:
			raise ValueError(f"A value for '{cls.__name__}' is required.") from None
	else:

		if cls.__name__ in raw_config_vars:
			return raw_config_vars[cls.__name__]
		else:
			if isinstance(cls.default, Callable):  # type: ignore
				return copy.deepcopy(cls.default(raw_config_vars))
			else:
				return copy.deepcopy(cls.default)


#: Mapping of Python types to their YAML equivalents.
yaml_type_lookup = {
		str: "String",
		int: "Integer",
		float: "Float",
		bool: "Boolean",
		list: "Sequence",
		dict: "Mapping",
		Any: "anything",
		}


def check_type(left: Type, *right: Type):
	return left in right or get_origin(left) in right


def get_yaml_type(type_: Type) -> str:
	r"""
	Get the YAML type that corresponds to the given Python type.

	:param type\_:
	"""

	if type_ in yaml_type_lookup:
		return yaml_type_lookup[type_]

	elif get_origin(type_) is Union:
		dtype = " or ".join(yaml_type_lookup[x] for x in type_.__args__)
		return dtype

	elif check_type(type_, list, List):
		args = get_args(type_)

		inner_types: typing.Iterable[str]

		if args:
			inner_types = (get_yaml_type(x) for x in args if not isinstance(x, TypeVar))
		else:
			inner_types = ()

		dtype = " or ".join(inner_types)
		if dtype:
			return f"Sequence of {dtype}"
		else:
			return "Sequence"

	elif check_type(type_, dict, Dict):
		args = get_args(type_)
		if not args or any(isinstance(t, TypeVar) for t in args):
			return "Mapping"
		else:
			dtype = " to ".join(get_yaml_type(x) for x in get_args(type_))
			return f"Mapping of {dtype}"

	elif is_literal_type(type_):
		types = [y for y in get_literal_values(type_)]
		return " or ".join(repr(x) for x in types)

	elif isinstance(type_, EnumMeta):
		return " or ".join([repr(x._value_) for x in type_])

	else:
		return str(type_)


basic_schema = MappingProxyType({
		"$schema": "http://json-schema.org/draft-07/schema",
		"type": "object",
		})


def make_schema(*configuration_variables: "ConfigVarMeta") -> Dict[str, Any]:
	"""
	Create a ``JSON`` schema from a list of :class:`~configconfig.configvar.ConfigVar` classes.

	:param configuration_variables:

	:return: Dictionary representation of the ``JSON`` schema.
	"""

	schema = {
			**basic_schema,
			"properties": {},
			"required": [],
			"additionalProperties": False,
			}

	for var in configuration_variables:
		schema = var.get_schema_entry(schema)

	return schema


def check_union(obj: Any, dtype: Union["GenericAliasType", "UnionType"]) -> bool:
	r"""
	Check if the type of ``obj`` is one of the types in a :class:`typing.Union`, :class:`typing.List` etc.

	:param obj:
	:param dtype:
	:type dtype: :py:obj:`~typing.Union`\, :class:`~typing.List`\, etc.
	"""

	args = dtype.__args__

	if float in args and int not in args:
		args = (*dtype.__args__, int)

	return isinstance(obj, args)


def get_json_type(type_: Type) -> Dict[str, Union[str, List, Dict]]:
	r"""
	Get the type for the JSON schema that corresponds to the given Python type.

	:param type\_:
	"""

	if type_ in json_type_lookup:
		return {"type": json_type_lookup[type_]}

	elif get_origin(type_) is Union:
		return {"type": [get_json_type(t)["type"] for t in type_.__args__]}

	elif check_type(type_, list, List):
		args = get_args(type_)

		if args:
			items = get_json_type(args[0])

			if items is NotImplemented:
				return {"type": "array"}
			elif "type" in items:
				return {"type": "array", "items": items}
			elif "enum" in items:
				return {"type": "array", "items": items}
			else:
				return {"type": "array"}

		return {"type": "array"}

	elif check_type(type_, dict, Dict):
		return {"type": "object"}

	elif check_type(type_, Literal) or is_literal_type(type_):  # type: ignore
		return {"enum": [x for x in get_literal_values(type_)]}

	elif isinstance(type_, EnumMeta):
		return {"enum": [x._value_ for x in type_]}

	elif type_ is bool:
		return {"type": ["boolean", "string"]}

	else:
		return NotImplemented


#: Mapping of Python types to their JSON equivalents.
json_type_lookup = {
		str: "string",
		int: "number",
		float: "number",
		dict: "object",
		list: "array",
		}


def get_literal_values(literal: Literal) -> typing.Tuple[Any]:  # type: ignore
	"""
	Returns a tuple of permitted values for a :class:`typing.Literal`.

	.. versionadded:: 0.3.0

	:param literal:
	"""

	if sys.version_info < (3, 7):
		return literal.__values__
	else:
		return literal.__args__


RawConfigVarsType = Dict[str, Any]



#!/usr/bin/env python3
#
#  validator.py
"""
Validate values obtained from the ``YAML`` file and coerce into the appropriate return types.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  validate based on https://github.com/yaccob/ytools
#  Copyright (c) Jakob Stemberger <yaccob@gmx.net>
#  Apache 2.0 Licensed
#

# stdlib
import pathlib
from typing import Any, Dict, Iterable, List, Optional, Union

# 3rd party
import jsonschema  # type: ignore
from domdf_python_tools.typing import PathLike
from domdf_python_tools.utils import strtobool
from ruamel.yaml import YAML
from typing_extensions import NoReturn
from typing_inspect import get_origin, is_literal_type  # type: ignore

# this package
from configconfig.metaclass import ConfigVarMeta
from configconfig.utils import RawConfigVarsType, check_union, get_literal_values, optional_getter

__all__ = ["Validator", "validate_files"]


class Validator:
	"""
	Methods are named ``visit_<type>``.
	"""

	def __init__(self, config_var: ConfigVarMeta):
		self.config_var = config_var

	_dtypes = {
			str: "str",
			int: "int",
			float: "float",
			bool: "bool",
			}

	def validate(self, raw_config_vars: Optional[RawConfigVarsType] = None) -> Any:
		"""
		Validate the configuration value.

		:param raw_config_vars:

		:returns: The validated value.
		"""

		if raw_config_vars is None:
			raw_config_vars = {}

		if self.config_var.rtype is None:
			self.config_var.rtype = self.config_var.dtype

		if self.config_var.dtype in self._dtypes:
			return getattr(self, f"visit_{self._dtypes[self.config_var.dtype]}")(raw_config_vars)

		elif get_origin(self.config_var.dtype) in {list, List}:
			return self.visit_list(raw_config_vars)

		elif get_origin(self.config_var.dtype) in {dict, Dict}:
			return self.visit_dict(raw_config_vars)

		elif get_origin(self.config_var.dtype) is Union:
			return self.visit_union(raw_config_vars)

		elif is_literal_type(self.config_var.dtype):
			return self.visit_literal(raw_config_vars)

		else:
			self.unknown_type()

	def _visit_str_number(self, raw_config_vars: RawConfigVarsType) -> Union[str, int, float]:
		obj = optional_getter(raw_config_vars, self.config_var, self.config_var.required)

		if not isinstance(obj, self.config_var.dtype):
			raise ValueError(f"'{self.config_var.__name__}' must be a {self.config_var.dtype}") from None

		return obj

	def visit_str(self, raw_config_vars: RawConfigVarsType) -> str:
		"""
		Used to validate and convert :class:`str` values.

		:param raw_config_vars:
		"""

		return self.config_var.rtype(self._visit_str_number(raw_config_vars))

	def visit_int(self, raw_config_vars: RawConfigVarsType) -> int:
		"""
		Used to validate and convert :class:`int` values.

		:param raw_config_vars:
		"""

		return self.config_var.rtype(self._visit_str_number(raw_config_vars))

	def visit_float(self, raw_config_vars: RawConfigVarsType) -> float:
		"""
		Used to validate and convert :class:`float` values.

		:param raw_config_vars:
		"""

		return self.config_var.rtype(self._visit_str_number(raw_config_vars))

	def visit_bool(self, raw_config_vars: RawConfigVarsType) -> bool:
		"""
		Used to validate and convert :class:`bool` values.

		:param raw_config_vars:
		"""

		obj = optional_getter(raw_config_vars, self.config_var, self.config_var.required)

		if not isinstance(obj, (int, bool, str)):
			raise ValueError(
					f"'{self.config_var.__name__}' must be one of {(int, bool, str)}, "
					f"not {type(obj)}"
					) from None

		return self.config_var.rtype(strtobool(obj))

	def visit_list(self, raw_config_vars: RawConfigVarsType) -> List:
		"""
		Used to validate and convert :class:`list` values.

		:param raw_config_vars:
		"""

		# Lists of strings, numbers, Unions and Literals
		buf = []

		data = optional_getter(raw_config_vars, self.config_var, self.config_var.required)
		if isinstance(data, str) or not isinstance(data, Iterable):
			raise ValueError(
					f"'{self.config_var.__name__}' must be a List of {self.config_var.dtype.__args__[0]}"
					) from None

		if get_origin(self.config_var.dtype.__args__[0]) is Union:
			for obj in data:
				if not check_union(obj, self.config_var.dtype.__args__[0]):
					raise ValueError(
							f"'{self.config_var.__name__}' must be a "
							f"List of {self.config_var.dtype.__args__[0]}"
							) from None

		elif is_literal_type(self.config_var.dtype.__args__[0]):
			for obj in data:
				# if isinstance(obj, str):
				# 	obj = obj.lower()
				if obj not in get_literal_values(self.config_var.dtype.__args__[0]):
					raise ValueError(
							f"Elements of '{self.config_var.__name__}' must be "
							f"one of {get_literal_values(self.config_var.dtype.__args__[0])}"
							) from None
		else:
			for obj in data:
				if not check_union(obj, self.config_var.dtype):
					raise ValueError(
							f"'{self.config_var.__name__}' must be a List of {self.config_var.dtype.__args__[0]}"
							) from None

		try:
			for obj in data:
				if self.config_var.rtype.__args__[0] in {int, str, float, bool}:
					buf.append(self.config_var.rtype.__args__[0](obj))
				else:
					buf.append(obj)

			return buf

		except ValueError:
			raise ValueError(
					f"Values in '{self.config_var.__name__}' must be {self.config_var.rtype.__args__[0]}"
					) from None

	def visit_dict(self, raw_config_vars: RawConfigVarsType) -> Dict:
		"""
		Used to validate and convert :class:`dict` values.

		:param raw_config_vars:
		"""

		# Dict[str, str]
		if self.config_var.dtype == Dict[str, str]:
			obj = optional_getter(raw_config_vars, self.config_var, self.config_var.required)
			if not isinstance(obj, dict):
				raise ValueError(f"'{self.config_var.__name__}' must be a dictionary") from None

			return {str(k): str(v) for k, v in obj.items()}

		# Dict[str, Any]
		elif self.config_var.dtype == Dict[str, Any]:
			obj = optional_getter(raw_config_vars, self.config_var, self.config_var.required)
			if not isinstance(obj, dict):
				raise ValueError(f"'{self.config_var.__name__}' must be a dictionary") from None

			return obj

		# Dict[str, List[str]
		elif self.config_var.dtype == Dict[str, List[str]]:
			obj = optional_getter(raw_config_vars, self.config_var, self.config_var.required)
			if not isinstance(obj, dict):
				raise ValueError(f"'{self.config_var.__name__}' must be a dictionary") from None

			return {str(k): [str(i) for i in v] for k, v in obj.items()}

		else:
			self.unknown_type()

	def visit_union(self, raw_config_vars: RawConfigVarsType) -> Any:
		"""
		Used to validate and convert :class:`typing.Union` values.

		:param raw_config_vars:
		"""

		obj = optional_getter(raw_config_vars, self.config_var, self.config_var.required)
		if not check_union(obj, self.config_var.dtype):
			raise ValueError(
					f"'{self.config_var.__name__}' must be one of {self.config_var.dtype.__args__}, "
					f"not {type(obj)}"
					) from None

		try:
			return self.config_var.rtype(obj)
		except ValueError:
			raise ValueError(
					f"'{self.config_var.__name__}' must be {self.config_var.rtype.__args__}, "
					f"not {type(obj)}"
					) from None

	def visit_literal(self, raw_config_vars: RawConfigVarsType) -> Any:
		"""
		Used to validate and convert :class:`typing.Literal` values.

		:param raw_config_vars:
		"""

		obj = optional_getter(raw_config_vars, self.config_var, self.config_var.required)
		# if isinstance(obj, str):
		# 	obj = obj.lower()
		if obj not in get_literal_values(self.config_var.dtype):
			raise ValueError(
					f"'{self.config_var.__name__}' must be one of {get_literal_values(self.config_var.dtype)}"
					) from None

		return obj

	def unknown_type(self) -> NoReturn:
		"""
		Called when the desired type has no visitor.
		"""

		print(self.config_var)
		print(self.config_var.dtype)
		print(get_origin(self.config_var.dtype))
		raise NotImplementedError


def validate_files(
		schemafile: PathLike,
		*datafiles: PathLike,
		encoding: str = "utf-8",
		) -> None:
	r"""
	Validate the given datafiles against the given schema.

	:param schemafile: The ``json`` or ``yaml`` formatted schema to validate with.
	:param \*datafiles: The ``json`` or ``yaml`` files to validate.
	:param encoding: Encoding to open the files with.

	.. versionadded:: 0.4.0
	"""

	schemafile = pathlib.Path(schemafile)

	yaml = YAML(typ="safe", pure=True)
	schema = yaml.load(schemafile.read_text(encoding=encoding))

	for filename in datafiles:
		for document in yaml.load_all(pathlib.Path(filename).read_text(encoding=encoding)):
			try:
				jsonschema.validate(document, schema, format_checker=jsonschema.FormatChecker())
			except jsonschema.exceptions.ValidationError as e:
				e.filename = str(filename)
				raise e



#!/usr/bin/env python3
#
#  testing.py
"""
Helpers for testing :class:`~configconfig.configvar.ConfigVar`.

.. extras-require:: testing
	:pyproject:

.. versionadded:: 0.2.0
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from abc import ABC
from typing import Any, Dict, List, Type

# 3rd party
import pytest  # nodep

# this package
from configconfig.configvar import ConfigVar

__all__ = [
		"ConfigVarTest",
		"NotIntTest",
		"NotBoolTest",
		"NotStrTest",
		"ListTest",
		"DirectoryTest",
		"BoolTrueTest",
		"BoolFalseTest",
		"RequiredStringTest",
		"OptionalStringTest",
		"EnumTest",
		"DictTest",
		"test_list_int",
		"test_list_str",
		]

test_list_int = [1, 2, 3, 4]
test_list_str = ['a', 'b', 'c', 'd']


class ConfigVarTest(ABC):
	r"""
	Base class for tests of :class:`~configconfig.configvar.ConfigVar`\s.
	"""

	#: The :class:`~configconfig.configvar.ConfigVar` under test.
	config_var: Type[ConfigVar]


class NotIntTest(ConfigVarTest):
	r"""
	Mixin to add tests for :class:`~configconfig.configvar.ConfigVar`\s
	that can't be integers.
	"""  # noqa: D400

	def test_error_int(self):
		"""
		Checks that the :class:`~configconfig.configvar.ConfigVar`
		raises a :class:`ValueError` when passed an :class:`int`.
		"""  # noqa: D400

		with pytest.raises(ValueError):  # noqa: PT011
			self.config_var.get({self.config_var.__name__: 1234})


class NotBoolTest(ConfigVarTest):
	r"""
	Mixin to add tests for :class:`~configconfig.configvar.ConfigVar`\s
	that can't be boolean values.
	"""  # noqa: D400

	def test_error_bool(self):
		"""
		Checks that the :class:`~configconfig.configvar.ConfigVar`
		raises a :class:`ValueError` when passed a :class:`bool`.
		"""  # noqa: D400

		with pytest.raises(ValueError):  # noqa: PT011
			self.config_var.get({self.config_var.__name__: True})


class NotStrTest(ConfigVarTest):
	r"""
	Mixin to add tests for :class:`~configconfig.configvar.ConfigVar`\s
	that can't be strings.
	"""  # noqa: D400

	def test_error_str(self):
		"""
		Checks that the :class:`~configconfig.configvar.ConfigVar`
		raises a :class:`ValueError` when passed a :class:`str`.
		"""  # noqa: D400

		with pytest.raises(ValueError):  # noqa: PT011
			self.config_var.get({self.config_var.__name__: "a string"})


class ListTest(NotStrTest, NotBoolTest, NotIntTest, ConfigVarTest):
	"""
	Test for list configuration values.
	"""

	#: A value that is valid and should be returned unchanged.
	test_value: List[str]

	#: The default value that should be returned when no valid is given.
	default_value: List[str] = []

	different_key_value: Dict[str, Any] = {"username": "domdfcoding"}
	r"""
	A dictionary containing one or more keys that are not the keys
	used by the :class:`~configconfig.configvar.ConfigVar`
	"""

	def test_success(self):
		"""
		Checks that the :class:`~configconfig.configvar.ConfigVar` can correctly parse various :class:`list` values.
		"""

		assert self.config_var.get({self.config_var.__name__: self.test_value}) == self.test_value
		assert self.config_var.get({self.config_var.__name__: []}) == []
		assert self.config_var.get(self.different_key_value) == self.default_value
		assert self.config_var.get() == self.default_value
		assert self.config_var.get({}) == self.default_value


class DirectoryTest(NotBoolTest, NotIntTest, ConfigVarTest):
	"""
	Test for configuration values which represent directories.
	"""

	#: A value that is valid and should be returned unchanged.
	test_value: str

	#: The default value that should be returned when no valid is given.
	default_value: str

	different_key_value: Dict[str, Any] = {"username": "domdfcoding"}
	r"""
	A dictionary containing one or more keys that are not the keys
	used by the :class:`~configconfig.configvar.ConfigVar`
	"""

	def test_success(self):
		"""
		Checks that the :class:`~configconfig.configvar.ConfigVar` can correctly parse various directory values.
		"""

		assert self.config_var.get({self.config_var.__name__: self.test_value}) == self.test_value
		assert self.config_var.get(self.different_key_value) == self.default_value
		assert self.config_var.get() == self.default_value
		assert self.config_var.get({}) == self.default_value

	def test_error_list_int(self):
		"""
		Checks that the :class:`~configconfig.configvar.ConfigVar`
		raises a :class:`ValueError` when passed a :class:`str`.
		"""  # noqa: D400

		with pytest.raises(ValueError, match="'.*' must be a <class 'str'>"):
			self.config_var.get({self.config_var.__name__: test_list_int})

	def test_error_list_str(self):
		"""
		Checks that the :class:`~configconfig.configvar.ConfigVar`
		raises a :class:`ValueError` when passed a :class:`str`.
		"""  # noqa: D400

		with pytest.raises(ValueError):  # noqa: PT011
			self.config_var.get({self.config_var.__name__: test_list_str})


class BoolTrueTest(ConfigVarTest):
	"""
	Test for boolean configuration values which default to :py:obj:`True`.
	"""

	different_key_value: Dict[str, Any] = {"username": "domdfcoding"}
	r"""
	A dictionary containing one or more keys that are not the keys
	used by the :class:`~configconfig.configvar.ConfigVar`
	"""

	@property
	def true_values(self) -> List[Dict[str, Any]]:
		"""
		A list of values which should be considered :py:obj:`True`
		by the :class:`~configconfig.configvar.ConfigVar`.
		"""  # noqa: D400

		return [
				{self.config_var.__name__: True},
				{self.config_var.__name__: 1},
				{self.config_var.__name__: 200},
				{self.config_var.__name__: -1},
				{self.config_var.__name__: "True"},
				self.different_key_value,
				{},
				]

	@property
	def false_values(self) -> List[Dict[str, Any]]:
		"""
		A list of values which should be considered :py:obj:`False`
		by the :class:`~configconfig.configvar.ConfigVar`.
		"""  # noqa: D400

		return [
				{self.config_var.__name__: 0},
				{self.config_var.__name__: False},
				{self.config_var.__name__: "False"},
				]

	def test_empty_get(self):  # noqa: D102
		assert self.config_var.get()

	def test_true(self):  # noqa: D102
		for true_value in self.true_values:
			assert self.config_var.get(true_value)

	def test_false(self):  # noqa: D102
		for false_value in self.false_values:
			assert not self.config_var.get(false_value)

	@property
	def wrong_values(self) -> List[Dict[str, Any]]:
		"""
		A list of values which should are of the wrong type.
		"""

		return [
				{self.config_var.__name__: "a string"},
				{self.config_var.__name__: test_list_int},
				{self.config_var.__name__: test_list_str},
				]

	def test_errors(self):  # noqa: D102
		for wrong_value in self.wrong_values:
			with pytest.raises(ValueError):  # noqa: PT011
				self.config_var.get(wrong_value)


class BoolFalseTest(BoolTrueTest):
	"""
	Test for boolean configuration values which default to :py:obj:`False`.
	"""

	different_key_value: Dict[str, Any] = {"username": "domdfcoding"}
	r"""
	A dictionary containing one or more keys that are not the keys
	used by the :class:`~configconfig.configvar.ConfigVar`
	"""

	@property
	def true_values(self) -> List[Dict[str, Any]]:  # noqa: D102
		return [
				{self.config_var.__name__: True},
				{self.config_var.__name__: 1},
				{self.config_var.__name__: 200},
				{self.config_var.__name__: -1},
				{self.config_var.__name__: "True"},
				]

	@property
	def false_values(self) -> List[Dict[str, Any]]:  # noqa: D102
		return [
				{self.config_var.__name__: 0},
				{self.config_var.__name__: False},
				{self.config_var.__name__: "False"},
				self.different_key_value,
				{},
				]

	def test_empty_get(self):  # noqa: D102
		assert not self.config_var.get()


class RequiredStringTest(ConfigVarTest):
	"""
	Test for string configuration values which are required.
	"""

	#: A value that is valid and should be returned unchanged.
	test_value: str

	def test_empty_get(self):  # noqa: D102
		with pytest.raises(ValueError):  # noqa: PT011
			self.config_var.get()

	def test_success(self):  # noqa: D102
		assert self.config_var.get({self.config_var.__name__: self.test_value}) == self.test_value

	@property
	def wrong_values(self) -> List[Dict[str, Any]]:
		"""
		A list of values which should are of the wrong type.
		"""

		return [
				{self.config_var.__name__: 1234},
				{self.config_var.__name__: True},
				{self.config_var.__name__: test_list_int},
				{self.config_var.__name__: test_list_str},
				]

	def test_errors(self):  # noqa: D102
		for wrong_value in self.wrong_values:
			with pytest.raises(ValueError):  # noqa: PT011
				self.config_var.get(wrong_value)


class OptionalStringTest(RequiredStringTest):
	"""
	Test for string configuration values which are optional.
	"""

	#: The default value that should be returned when no valid is given.
	default_value: str = ''

	different_key_value: Dict[str, Any] = {"sphinx_html_theme": "alabaster"}
	r"""
	A dictionary containing one or more keys that are not the keys
	used by the :class:`~configconfig.configvar.ConfigVar`
	"""

	def test_empty_get(self):  # noqa: D102
		assert self.config_var.get() == self.default_value
		assert self.config_var.get({}) == self.default_value

	def test_success(self):  # noqa: D102
		assert self.config_var.get({self.config_var.__name__: ''}) == ''
		assert self.config_var.get(self.different_key_value) == self.default_value
		super().test_success()

	@property
	def wrong_values(self) -> List[Dict[str, Any]]:  # noqa: D102
		return [
				{self.config_var.__name__: 1234},
				{self.config_var.__name__: True},
				{self.config_var.__name__: test_list_int},
				{self.config_var.__name__: test_list_str},
				]

	def test_errors(self):  # noqa: D102
		for wrong_value in self.wrong_values:
			with pytest.raises(ValueError):  # noqa: PT011
				self.config_var.get(wrong_value)


class EnumTest(RequiredStringTest):
	"""
	Test for :class:`~enum.Enum` configuration values.
	"""

	#: A list of values which are of the correct type but are invalid.
	non_enum_values: List[Any]

	#: The default value that should be returned when no valid is given.
	default_value: str

	def test_empty_get(self):  # noqa: D102
		assert self.config_var.get() == self.default_value
		assert self.config_var.get({}) == self.default_value

	def test_non_enum(self):  # noqa: D102
		for non_enum in self.non_enum_values:
			with pytest.raises(ValueError):  # noqa: PT011
				self.config_var.get({self.config_var.__name__: non_enum})

	def test_errors(self):  # noqa: D102
		wrong_values: List[Dict[str, Any]] = [
				{self.config_var.__name__: 1234},
				{self.config_var.__name__: True},
				{self.config_var.__name__: test_list_int},
				{self.config_var.__name__: test_list_str},
				]
		for wrong_value in wrong_values:
			with pytest.raises(ValueError):  # noqa: PT011
				self.config_var.get(wrong_value)


class DictTest(NotStrTest, NotBoolTest, NotIntTest, ConfigVarTest):
	"""
	Test for dictionary configuration values.
	"""

	#: A value that is valid and should be returned unchanged.
	test_value: Dict[str, Any]

	#: The default value that should be returned when no valid is given.
	default_value: Dict[str, Any] = {}

	different_key_value: Dict[str, Any] = {"sphinx_html_theme": "alabaster"}
	r"""
	A dictionary containing one or more keys that are not the keys
	used by the :class:`~configconfig.configvar.ConfigVar`
	"""

	def test_success(self):  # noqa: D102
		assert self.config_var.get({self.config_var.__name__: self.test_value}) == self.test_value
		assert self.config_var.get({self.config_var.__name__: {}}) == {}
		assert self.config_var.get(self.different_key_value) == self.default_value
		assert self.config_var.get() == self.default_value
		assert self.config_var.get({}) == self.default_value

	def test_error_list_int(self):  # noqa: D102
		with pytest.raises(ValueError):  # noqa: PT011
			self.config_var.get({self.config_var.__name__: test_list_int})



#!/usr/bin/env python3
#
#  metaclass.py
"""
Metaclass for configuration values.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Dict, Mapping, Optional, Type, cast

# this package
from configconfig.utils import basic_schema, get_json_type

__all__ = ["ConfigVarMeta"]

if TYPE_CHECKING:
	# this package
	from configconfig.configvar import ConfigVar


class ConfigVarMeta(type):
	"""
	Metaclass for configuration values.
	"""

	dtype: Type
	rtype: Type
	required: bool
	default: Any
	validator: Callable
	category: str
	__name__: str

	def __new__(cls, name: str, bases, dct: Dict):  # noqa: D102
		x = cast("ConfigVar", super().__new__(cls, name, bases, dct))

		def get(name, default):
			return dct.get(name, getattr(x, name, default))

		x.dtype = get("dtype", Any)

		if "rtype" in dct:
			x.rtype = dct["rtype"]
		elif getattr(x, "rtype", Any) != Any:
			pass
		else:
			x.rtype = x.dtype

		x.required = get("required", False)
		x.default = get("default", '')
		x.validator = get("validator", lambda y: y)  # type: ignore
		x.category = get("category", "other")
		x.__name__ = dct.get("name", dct.get("__name__", x.__name__))

		return x

	def get_schema_entry(cls, schema: Optional[Dict] = None) -> Dict[str, Any]:
		"""
		Returns the JSON schema entry for this configuration value.

		:param schema:

		:return: Dictionary representation of the JSON schema.
		"""

		if schema is None:
			schema = {
					**basic_schema,
					"properties": {},
					"required": [],
					}

		dtype = get_json_type(cls.dtype)
		if dtype is NotImplemented:
			raise NotImplementedError(cls.__name__, cls.dtype)
		else:
			schema["properties"][cls.__name__] = dtype

		if cls.required:
			schema["required"].append(cls.__name__)

		for line in (cls.__doc__ or '').split("\n\n"):
			line = ' '.join([p.strip() for p in line.split('\n') if p.strip()])
			if line:
				schema["properties"][cls.__name__]["description"] = line
				break

		return schema

	@property
	def schema_entry(cls) -> Dict[str, Any]:  # noqa: D102
		return cls.get_schema_entry()

	def __call__(cls, raw_config_vars: Dict[str, Any]) -> Any:  # type: ignore  # noqa: D102
		"""
		Alias for :meth:`ConfigVar.get <.ConfigVar.get>`.

		Returns the value of the :class:`~configconfig.configvar.ConfigVar`.

		:param raw_config_vars: Dictionary to obtain the value from.

		:rtype: See the :attr:`ConfigVar.rtype <.ConfigVar.rtype>` attribute.
		"""

		return cls.get(raw_config_vars)

	@abstractmethod
	def get(cls, raw_config_vars: Mapping[str, Any]):  # pragma: no cover  # noqa: D102
		return NotImplemented

	def __repr__(self) -> str:
		"""
		Return a string representation of the :class:`~.ConfigVarMeta` class.
		"""

		return f"<ConfigVar {self.__name__!r}>"



#!/usr/bin/env python3
#
#  autoconfig.py
"""
A Sphinx directive for documenting YAML configuration values.

Provides the :rst:dir:`autoconfig` directive to document configuration
values automatically, the :rst:dir:`conf` directive to document them manually,
and the :rst:role:`conf` role to link to a :rst:dir:`conf` directive.

.. extras-require:: sphinx
	:pyproject:


Usage
---------

.. rst:directive:: autoconfig

	Directive to automatically document an YAML configuration value.

	Takes a single argument, either the fully qualified name of the :class:`~.ConfigVar` object,
	or the name of the module if the ``:category:`` option is given.

	.. rst:directive:option:: category
		:type: string

		(optional) The category of options to document.


.. rst:directive:: conf

	Directive to document an YAML configuration value.


.. rst:role:: conf

	Role to add a cross-reference to a :rst:dir:`conf` or :rst:dir:`autoconfig` directive.


.. latex:clearpage::

API Reference
---------------
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  ``.. conf::`` directive and ``:conf:`` role based on Tox's documentation.
#  From https://github.com/tox-dev/tox/blob/master/docs/conf.py
#  MIT Licensed
#

# stdlib
import warnings
from typing import Any, Dict, List, Sequence, Type

# 3rd party
from docutils import nodes  # nodep
from docutils.parsers.rst.directives import unchanged  # nodep
from docutils.statemachine import StringList  # nodep
from sphinx import addnodes  # nodep
from sphinx.application import Sphinx  # nodep
from sphinx.environment import BuildEnvironment  # nodep
from sphinx.ext.autodoc.importer import import_module, import_object  # nodep
from sphinx.util.docutils import SphinxDirective  # nodep
from sphinx_toolbox.utils import Purger  # nodep

# this package
from configconfig import __version__
from configconfig.configvar import ConfigVar
from configconfig.metaclass import ConfigVarMeta

__all__ = ["AutoConfigDirective", "conf_node_purger", "parse_conf_node", "setup"]

conf_node_purger = Purger("all_conf_nodes")


class AutoConfigDirective(SphinxDirective):
	"""
	Sphinx directive to automatically document an YAML configuration value.
	"""

	has_content: bool = True
	required_arguments: int = 1
	# the fully qualified name of the ConfigVar object,
	# or the name of the module if :category: given
	option_spec = {"category": unchanged}

	def run(self) -> Sequence[nodes.Node]:  # type: ignore
		"""
		Process the content of the directive.
		"""

		config_var: str = self.arguments[0]

		if "category" in self.options:
			node_list = []
			module = import_module(config_var)

			if hasattr(module, "__all__"):
				module_all = module.__all__
			else:
				module_all = module.__dict__

			category = self.options["category"]

			# TODO: category header
			# 			header = f"""\
			# ={'='*len(category)}
			# {category.capitalize()}
			# ={'='*len(category)}
			#
			# """
			# 			content = header.replace("\t", "    ")
			# 			view = ViewList(content.split("\n"))
			# 			config_node = nodes.paragraph(rawsource=content)
			# 			self.state.nested_parse(view, self.content_offset, config_node)

			for class_ in module_all:
				var_obj: Type[ConfigVar] = getattr(module, class_)

				if not (isinstance(var_obj, ConfigVarMeta) and issubclass(var_obj, ConfigVar)):
					continue  # pragma: no cover
				elif var_obj.category == category:
					node_list.append(self.document_config_var(var_obj))

			return node_list

		else:
			module_name, class_ = config_var.rsplit('.', 1)
			var_obj = import_object(module_name, [class_])[3]
			if not issubclass(var_obj, ConfigVar):
				warnings.warn("'autoconfig' can only be used with 'ConfigVar' subclasses.")
				return []

			return [self.document_config_var(var_obj)]

	def document_config_var(self, var_obj: Type[ConfigVar]) -> nodes.paragraph:
		"""
		Document the given configuration value.

		:param var_obj:
		"""

		docstring = var_obj.make_documentation()

		targetid = f'autoconfig-{self.env.new_serialno("autoconfig"):d}'
		targetnode = nodes.section(ids=[targetid])

		content = docstring.replace('\t', "    ")
		view = StringList(content.split('\n'))
		config_node = nodes.paragraph(rawsource=content)
		self.state.nested_parse(view, self.content_offset, config_node)

		conf_node_purger.add_node(self.env, config_node, targetnode, self.lineno)

		return config_node


def parse_conf_node(env: BuildEnvironment, text: str, node: addnodes.desc_signature) -> str:
	"""
	Parse the content of a :rst:dir:`conf` directive.

	:param env: The Sphinx build environment.
	:param text: The content of the directive.
	:param node: The docutils node class.
	"""

	args = text.split('^')
	name = args[0].strip()

	node += addnodes.literal_strong(name, name)

	if len(args) > 2:
		default = f"={args[2].strip()}"
		node += nodes.literal(text=default)

	if len(args) > 1:
		content = f"({args[1].strip()})"
		node += addnodes.compact_paragraph(text=content)

	return name  # this will be the link


def _get_outdated(
		app: Sphinx,
		env: BuildEnvironment,
		added: List[str],
		changed: List[str],
		removed: List[str],
		) -> List[str]:
	if not hasattr(env, conf_node_purger.attr_name):
		return []

	return [todo["docname"] for todo in getattr(env, conf_node_purger.attr_name)]


def setup(app: Sphinx) -> Dict[str, Any]:
	"""
	Setup Sphinx Extension.

	:param app: The Sphinx app
	"""

	app.add_directive("autoconfig", AutoConfigDirective)
	app.connect("env-purge-doc", conf_node_purger.purge_nodes)
	app.connect("env-get-outdated", _get_outdated)

	app.add_object_type(
			directivename="conf",
			rolename="conf",
			objname="configuration value",
			indextemplate="pair: %s; configuration value",
			parse_node=parse_conf_node,
			)

	return {
			"version": __version__,
			"parallel_read_safe": True,
			"parallel_write_safe": True,
			}
